/*****************************************************************************
 * kate_categories.h : maps well known category tags to translated strings.
 *****************************************************************************
 * Copyright (C) 2009 ogg.k.ogg.k@googlemail.com
 * $Id: f3118dfa3606ee2909274441a78b9081c34048f2 $
 *
 * Authors: ogg.k.ogg.k@googlemail.com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston MA 02110-1301, USA.
 *****************************************************************************/

#ifndef KATE_CATEGORIES_H_
#define KATE_CATEGORIES_H_ 1

const char *FindKateCategoryName( const char *psz_tag );

#endif

