style "videolan-notebook"
{
  GtkNotebook::inner_left_border = 0
  GtkNotebook::inner_right_border = 0
  GtkNotebook::tab-overlap = 0
  GtkNotebook::arrow-spacing = 6
  GtkNotebook::label-padding = 12
  GtkWidget::scroll-arrow-hlength = 20
  GtkWidget::scroll-arrow-vlength = 30

  engine "sapwood"
  {
    image
    {
      function = EXTENSION
      state = ACTIVE
      file = "vlc_left_tab_passive.png"
      border = { 12, 12, 12, 12 }
      gap_side = RIGHT
    }
    image
    {
      function = EXTENSION
      file = "vlc_left_tab_active.png"
      border = { 12, 12, 12, 12 }
      gap_side = RIGHT
    }
  }
}
class "GtkNotebook" style "videolan-notebook"

style "videolan"
{
  stock["vlc-next"] = {{ "next.png" }}
  stock["vlc-previous"] = {{ "previous.png" }}
  stock["vlc-stop"] = {{ "stop.png" }}
  stock["vlc-play"] = {{ "play.png" }}
  stock["vlc-pause"] = {{ "pause.png" }}
  stock["vlc-playlist"] = {{ "playlist.png" }}
  stock["vlc"] = {{ "../../icons/hicolor/32x32/apps/vlc.png" }}
}
widget "*" style "videolan"
