/**
 * Pre-computed Diffie-Hellman-Merkel parameters.
 * $Id: d7e97c5d69c997e810f76f195bed590b12d707d9 $
 * Numbers are not copyrightable.
 *
 * If you feel like it, generate new ones:
 * # certtool --generate-dh-params --bits 2048
 */
static const unsigned char dh_params[] =
    "-----BEGIN DH PARAMETERS-----\n"
    "MIIBBwKCAQDJ9pqmqBy71hn5pA3QL1AiGB2JOKq2wfdRq3EQVdbOtPscXM6BXdm7"
    "NfBRUZIGT47oPNgCOhVV33z9OfnMHCSiMoWFPZeT14Mdm5TQBBYA2H6tf0g2Fp4X"
    "qs7eeYheslzkg1V7U7PcdsyopiGI0FC4Heq+PxcyFOC0DuDUNiRq2Pk51gSUQURS"
    "dtTyiU5fEFUETcFN2FWuPuCdQpA0xmYnQwnTJKq75b1GAxRlp6XqTbWmXgBggwYk"
    "+O/oGpgVLlwZPquSbumbQsp4OU0Lk0hxqTR4Jd3XIeBuV/pc1zZaCH/9LIQT6aXn"
    "S1wHhOlCUKgPWlBtx82omKgyo8ebJwWjAgEG\n"
    "-----END DH PARAMETERS-----";
